;********************************************************************
;	MSBINCVT.ASM - subroutines to convert between the (obsolete)
;	Microsoft Binary Format and IEEE format numbers.
;
;	IMPORTANT NOTE: All these functions manipulate the relative
;	positions of the sign bit and exponent fields. Therefore, the
;	comments will often look like:
;
;	AX = EEEEEEEESMMMMMMM
;
;	Where we use:
;		M = Mantissa bit
;		E = Microsoft exponent bit
;		S = Sign bit
;		e = IEEE exponent bit
;
;	Therefore the comments often just show the result of the opcode
;	just executed.
;********************************************************************


.model MEMORY_MODEL

	include MACROS.INC			; @PopAll,@PushAll

.code


;--------------------------------------------------------------------
;	MSBtoIEEEshort$n - convert an old Microsoft single precision
;	float to IEEE short real format
;
;	INP:	DI - points to the MSB number
;	OUT:	CY - TRUE if exponent range error detected, FALSE o/w
;			the number is converted, in place, to IEEE format
;	USES:	flags
;
MSBtoIEEEshort$n proc

	push	ax
	mov 	ax, [di+2]			; AX = EEEEEEEESMMMMMMM
	sub 	ah, 2				; AX = eeeeeeeeSMMMMMMM fix exponent
	jna 	M2Is_error			; If exponent out of range, error
	rcl 	al, 1				; AX = eeeeeeeeMMMMMMM?,  CF = S
	rcr 	ax, 1				; AX = SeeeeeeeeMMMMMMM
	mov 	[di+2], ax			; Store result
	pop 	ax
	clc 						; Clear carry -- i.e. no errors
	ret

M2Is_error:						; Error detected -- don't convert
	pop 	ax
	stc 						; Set carry -- i.e. return error
	ret


MSBtoIEEEshort$n endp


;--------------------------------------------------------------------
;	IEEEtoMSBfloat$n - convert an IEEE short real to old Microsoft
;	single precision format
;
;	INP:	DI - points to the IEEE number to convert
;	OUT:	CY - TRUE if exponent range error detected, FALSE o/w
;			the number is converted, in place, to the old MS format
;	USES:	flags
;
IEEEtoMSBshort$n proc

	push	ax
	mov 	ax, [di+2]			; AX = SeeeeeeeeMMMMMMM
	rcl 	ax, 1				; AX = eeeeeeeeMMMMMMM?,  CF=S
	rcr 	al, 1				; AX = eeeeeeeeSMMMMMMM
	add 	ah, 2				; AX = EEEEEEEESMMMMMMM
	jc		I2Ms_error			; If exponent overflows -- error
	mov 	[di+2], ax			; Store result
	pop 	ax
	clc 						; Clear carry -- i.e. no errors
	ret

I2Ms_error: 					; Error detected -- don't convert
	pop 	ax
	stc 						; Set carry -- i.e. return error
	ret

IEEEtoMSBshort$n endp


;--------------------------------------------------------------------
;	MSBtoIEEElong$n - convert Microsoft Binary long precision to
;	IEEE long real format
;
;	INP:	DI - points to the MSB number
;	OUT:	CY - FALSE (for consistency--no errors are possible)
;			the number is converted, in place, to IEEE format
;	USES:	flags
;	NOTE:	No error possible (values ALWAYS fit)
;
MSBtoIEEElong$n proc

	@PushAll

	; Shift entire 64 bit number right 3 bits
	; At start: AX/DX/DI/SI = EEEEEEEESMMMMMMMMMMMM.....

	mov 	si, [di]			; Load AX:DX:BX:SI
	mov 	bx, [di+2]
	mov 	dx, [di+4]
	mov 	ax, [di+6]
	mov 	cx, 3				; # bits to shift
M2Iloop:
	shr 	al, 1				; Shift mantissa one bit right
	rcr 	dx, 1
	rcr 	bx, 1
	rcr 	si, 1
	loop	M2Iloop
	mov 	[di], si			; Store lower three words
	mov 	[di+2], bx
	mov 	[di+4], dx
	; Now AX = EEEEEEEE000SMMMM

	; Adjust exponent & swap around sign & exponent fields

	mov 	dl, al				; DL = 000SMMMM
	xchg	al, ah
	xor		ah, ah				; AX = 00000000EEEEEEEE
	add 	ax, 037eh			; AX = 00000eeeeeeeeeee
	test	dl, 10h 			; Check S, if on set new S in AX
	jz		M2Iskip
	or		ah, 08h
M2Iskip:						; AX = 0000Seeeeeeeeeee
	mov 	cl, 4
	shl 	ax, cl				; AX = Seeeeeeeeeee0000
	and 	dl, 0fh 			; DL = 0000MMMM
	or		al, dl				; AX = SeeeeeeeeeeeMMMM
	mov 	[di+6], ax			; Store the new upper word

	@PopAll
	clc 						; clear carry -- i.e. no errors
	ret

MSBtoIEEElong$n endp


;--------------------------------------------------------------------
;	IEEEtoMSBdouble$n - convert IEEE double to Old Microsoft format
;
;	INP:	DI - points to the MSB number
;	OUT:	CY - TRUE if exponent range error detected, FALSE o/w
;			the number is converted, in place, to Old Microsoft format
;	USES:	flags
;	NOTE:	No error checking performed
;
IEEEtoMSBlong$n proc

	@PushAll

	; Adjust exponent & swap around sign & exponent fields

	mov 	ax, [di+6]			; AX = SeeeeeeeeeeeMMMM
	mov 	dx, ax				; <keep a copy...>
	mov 	cl, 4
	shr 	ax, cl				; AX = 0000Seeeeeeeeeee
	and 	ah, 07h 			; AX = 00000eeeeeeeeeee
	sub 	ax, 037eh			; AX = 00000???EEEEEEEE
	jc		I2Ml_error			; Check exponent range for overflow
	cmp 	ax, 100h			; and underflow
	jnc		I2Ml_error
	mov 	[di+7], al			; Store the exponent
	mov 	ax, dx				; AX = SeeeeeeeeeeeMMMM again

	; Shift entire 64 bit number left 3 bits
	; At start: AX/DX/DI/SI = 000SeeeeeeeeMMMMMMMMMMMMMMMMMMMMMM.....

	mov 	si, [di]			; Load next three words--DX:BX:SI
	mov 	bx, [di+2]
	mov 	dx, [di+4]
	mov 	cx, 3				; # bits to shift
M2Iloop:
	shl 	si, 1				; Shift mantissa left 1 bit
	rcl 	bx, 1
	rcl 	dx, 1
	rcl 	al, 1
	loop	M2Iloop
	; Now AX = SeeeeeeeeMMMMMMM
	shl 	ax, 1				; AX = eeeeeeeeMMMMMMM0, CF = S
	rcr 	al, 1				; AX = eeeeeeeeSMMMMMMM
	mov 	[di], si			; Store the result
	mov 	[di+2], bx
	mov 	[di+4], dx
	mov 	[di+6], al			; (already stored exp at DI+7)

	@PopAll
	clc
	ret

I2Ml_error: 					; Error detected -- don't convert
	@PopAll
	stc 						; Set carry -- i.e. error detected
	ret

IEEEtoMSBlong$n endp

	end
